
from __future__ import annotations

import math
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
from scipy import sparse
from sksparse import cholmod

from .neighbours import Neighbours
from .utils import map_1d_2d_nx, map_2d_1d_nx_ny, append_element

if TYPE_CHECKING:
    from .grid import Grid

__all__ = ['Spde']

class Spde:
    """Represents a Stochastic Partial Differential Equation (SPDE) solver.

    Attributes
    ----------
    _v : int
        Order of the modified Bessel function of the second kind.
    _d : int
        Dimensions.

    Methods
    -------
    __init__(grid: Grid, sigma: float = 1, a: float = 50,
    dtype: type = np.float64) -> None
        Initialize the SPDE.
    revert_padding(Z_M: np.ndarray) -> np.ndarray
        Reverts the padding.
    create_system() -> sparse.csc_matrix
        Create the system.
    calc_element_terms(A: sparse.dok_matrix, i: int, j: int,
     neigh: Neighbours) -> None
        Calculate the elements of the matrix A.
    calc_derivatives(i: int, j: int, neigh: Neighbours) -> tuple
        Calculate the derivatives at a given point on the grid.
    simulate(n: int = 1, seed: int | None = None) -> np.ndarray
        Simulate the Stochastic Partial Differential Equation solver.

    """

    _v = 1
    _d = 2

    def __init__(self, grid: Grid, sigma: float = 1, a: float = 50,
                 dtype: type = np.float64) -> None:
        """Initialize the SPDE.

        Args:
        ----
            grid (Grid): The grid object to initialize the class.
            sigma (float): The sigma value for calculations. Default is 1.
            a (float): The a value for calculations. Default is 50.
            dtype (type): The data type for calculations. Default is np.float64.

        """
        self.grid = grid
        self.dtype = dtype

        alpha = self._v + self._d / 2

        self.k = 1 / a
        self.tau = ((sigma * (math.gamma(alpha) ** 0.5) *
                    ((4 * np.pi) ** (self._d / 4)) * (self.k**self._v)) /
                    math.gamma(self._v) ** 0.5)

        self.A, self.det_H = self.create_system()



    def revert_padding(self, Z_M: np.ndarray) -> np.ndarray:
        """Reverts the padding.

        Args:
        ----
            Z_M (np.ndarray): The input numpy array.

        Returns:
        -------
            np.ndarray: The numpy array with padding reverted.

        """
        if Z_M.ndim == 2:
            Z_M = Z_M[self.grid.pady:-self.grid.pady,
                      self.grid.padx:-self.grid.padx]
        else:
            Z_M = Z_M[self.grid.pady:-self.grid.pady,
                      self.grid.padx:-self.grid.padx,
                      :]
        return Z_M


    def create_system(self) -> sparse.csc_matrix:
        """Create the system.

        Returns
        -------
            sparse.csc_matrix: The sparse matrix representing the system.

        """
        map_2d_1d = partial(map_2d_1d_nx_ny, nx=self.grid._nx,
                                             ny=self.grid._ny)
        n = self.grid._nx*self.grid._ny

        vals, rows, cols = [], [], []
        det_H = np.ndarray(n, dtype=self.dtype)

        for i in range(self.grid._ny):
            for j in range(self.grid._nx):

                neigh = Neighbours(i, j, map_2d_1d)
                self.calc_element_terms(vals, rows, cols, det_H, i, j, neigh)

        A = sparse.csc_matrix((vals, (rows, cols)))

        return A, det_H

    def calc_element_terms(self, vals: list, rows: list, cols: list,
                           det_H: np.ndarray, 
                           i: int, j: int, neigh: Neighbours) -> None:
        """Calculate the elements of the matrix A.

        Args:
        ----
        A : sparse.dok_matrix
            The matrix to store the calculated elements.
        i : int
            The row index in the matrix.
        j : int
            The column index in the matrix.
        neigh : Neighbours
            Object containing information about neighboring elements in the
            grid.

        """
        a, b, c, d, e, p, q = self.calc_derivatives(i, j, neigh)

        append_element(vals, rows, cols, (2*q + 2*p) + self.k**2,
                       neigh.id, neigh.id)

        H_x = self.grid.anisotropy.H[i, j]
        det_H[neigh.id] = H_x[0,0]*H_x[1,1] - H_x[0,1]*H_x[1,0]

        if neigh.id_bot is not None:
            append_element(vals, rows, cols,- c - p - b,
                           neigh.id, neigh.id_bot)

        if neigh.id_bot_left is not None:
            append_element(vals, rows, cols,e,
                           neigh.id, neigh.id_bot_left)

        if neigh.id_left is not None:
            append_element(vals, rows, cols,- q + d + a,
                           neigh.id, neigh.id_left)

        if neigh.id_top_left is not None:
            append_element(vals, rows, cols, - e,
                            neigh.id, neigh.id_top_left)

        if neigh.id_top is not None:
            append_element(vals, rows, cols,- p + c + b,
                            neigh.id, neigh.id_top)

        if neigh.id_top_right is not None:
            append_element(vals, rows, cols,e,
                             neigh.id, neigh.id_top_right)

        if neigh.id_right is not None:
            append_element(vals, rows, cols,- d - q - a,
                            neigh.id, neigh.id_right)

        if neigh.id_bot_right is not None:
            append_element(vals, rows, cols,- e,
                           neigh.id, neigh.id_bot_right)


    def calc_derivatives(self, i: int, j: int, neigh: Neighbours) -> tuple:
        """Calculate the derivatives at a given point on the grid.

        Parameters
        ----------
        i : int
            The x-coordinate of the point.
        j : int
            The y-coordinate of the point.
        neigh : Neighbours
            Neighbouring points information.

        Returns
        -------
        tuple
            Tuple containing the calculated derivatives a, b, c, d, e, p, q.

        """
        H = self.grid.anisotropy.H
        dx = self.grid.dx
        dy = self.grid.dy

        if neigh.has('id_top') & neigh.has('id_bot'):
            a = (H[i + 1, j, 0, 1] - H[i - 1, j, 0, 1])/(4*dx*dy)
            c = (H[i  + 1, j, 0, 0] - H[i - 1, j, 0, 0])/(4*dy**2)
        else:
            a = 0.0
            c = 0.0

        if neigh.has('id_left') & neigh.has('id_right'):
            b = (H[i, j + 1, 1, 0] - H[i, j - 1, 1, 0])/(4*dx*dy)
            d = (H[i, j + 1, 1, 1] - H[i, j - 1, 1, 1])/(4*dx**2)
        else:
            b = 0.0
            d = 0.0

        e = (H[i, j, 0, 1] + H[i, j, 1, 0])/(4*dx*dy)

        p = H[i, j, 0, 0]/dy**2
        q = H[i, j, 1, 1]/dx**2

        return a, b, c, d, e, p, q


    def simulate(self, n: int = 1, seed: int | None = None) -> np.ndarray:
        """Simulate.

        Parameters
        ----------
        n : int, optional
            Number of simulations to generate. Default is 1.
        seed : int or None, optional
            Random seed for reproducibility. Default is None.

        Returns
        -------
        np.ndarray
            Simulation results in the shape (ny, nx, n).

        """
        rng = np.random.default_rng(seed=seed)

        Q_yy = (self.A.T @ self.A).tocsc() 

        try:
            factor = cholmod.cholesky(Q_yy)

        except cholmod.CholmodNotPositiveDefiniteError:
            msg = ('Anisotropy field is not positive definite, '
                   'check anisotropy.check_positiveness()')
            raise RuntimeError(msg)

        W = rng.standard_normal(size=(Q_yy.shape[0], n))

        Z = factor.solve_Lt(
                W
                *(self.tau**2 * (self.det_H**0.5)[factor.P() ,np.newaxis])**0.5
                ,
                use_LDLt_decomposition=False)

        idxs = np.argsort(factor.P())

        Z = Z[idxs, :]

        Z_M = Z.reshape((self.grid._ny, self.grid._nx, n))

        Z_M = self.revert_padding(Z_M)

        return Z_M


    def kriging(self, samples: np.ndarray) -> np.ndarray:
        """Solve for Simple kriging, based on the provided samples.

        Parameters
        ----------
        samples : np.ndarray
            numpy array of shape (n_samples, 3) containing the sample points
            with x, y coordinates and values.

        Returns
        -------
        np.ndarray
            numpy array of shape (ny, nx) representing the kriged 2D grid.

        """
        xv = samples[:,0].astype(int)
        yv = samples[:,1].astype(int)
        z_ = samples[:,2]

        n_samps = z_.shape[0]

        xv += self.grid.padx
        yv += self.grid.pady

        map_2d_1d = partial(map_2d_1d_nx_ny, nx=self.grid._nx,
                                             ny=self.grid._ny)
        map_1d_2d = partial(map_1d_2d_nx, nx=self.grid._nx)

        samples_idxs = []
        for i, j in zip(yv, xv):
            idx = map_2d_1d(i, j)
            samples_idxs.append(idx)


        target_idxs, i_grid, j_grid = [], [], []

        for idx in range(self.grid._nx * self.grid._ny):
            if idx not in samples_idxs:
                target_idxs.append(idx)
                i, j = map_1d_2d(idx)
                i_grid.append(i)
                j_grid.append(j)

        A_xx = self.A[samples_idxs, :][:, samples_idxs]
        A_yx = self.A[target_idxs, :][:, samples_idxs]
        A_yy = self.A[target_idxs, :][:, target_idxs]

        A_xy = A_yx.T

        A = sparse.vstack([sparse.hstack([A_xx, A_xy]),
                           sparse.hstack([A_yx, A_yy])])

        Q = (A.T @ A).tocsc()

        Q_yy = Q[n_samps:, n_samps:]
        Q_yx = Q[n_samps:, :n_samps]

        try:
            factor = cholmod.cholesky(Q_yy)

        except cholmod.CholmodNotPositiveDefiniteError:
            msg = ('Anisotropy field is not positive definite, '
                   'check anisotropy.check_positiveness()')
            raise RuntimeError(msg)

        Z = -factor.solve_A(Q_yx @ z_[:, np.newaxis])[:, 0]

        Z_M = np.empty((self.grid._ny, self.grid._nx))
        Z_M[i_grid, j_grid] = Z
        Z_M[yv, xv] = z_

        Z_M = self.revert_padding(Z_M)\
                  .squeeze()

        return Z_M






import re

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
import xml.etree.ElementTree as Et
from matplotlib.figure import Figure

from .utils import get_inkscape_namespace, get_inkscape_transform

__all__ = ['Anisotropy', 'anisotropy_from_svg', 'anisotropy_stationary']

class Anisotropy:
    """Represents an anisotropy object.

    Attributes
    ----------
        _x (np.ndarray): X-coordinates of vectors.
        _y (np.ndarray): Y-coordinates of vectors.
        _u (np.ndarray): Horizontal components of vectors.
        _v (np.ndarray): Vertical components of vectors.
        width (float): Width extent of the anisotropy.
        height (float): Height extent of the anisotropy.
        H (np.ndarray): Anisotropy tensor.
        V (np.ndarray): Vector field tensor.
        beta (float): Anisotropic local effect coefficient.
        gamma (float): Isotropic baseline effect coeficient.
        dtype (type): Data type for the fields tensors.

    """

    def __init__(self, x_u: np.ndarray, beta: float, gamma: float,
                 width: float = None, height: float = None, k: int = 3,
                 dtype: type = np.float64, *, scale: bool = True) -> None:
        """Initialize an Anisotropy object.

        Args:
        ----
            x_u (np.ndarray): Anisotropy data points of shape (p, 4). Each row
            is composed by  (x, y, u, v) where (x, y) coordinates with (u, v)
            directions.
            beta (float): Anisotropic local effect coefficient.
            gamma (float): Isotropic baseline effect coeficient.
            width (float): Width extent of the anisotropy.
            height (float): Height extent of the anisotropy.
            k (int): No of neighbours to consider during interpolation.
            dtype (type): Data type for the fields tensors.
            Defaults to np.float32.
            scale (bool, optional): Whether to scale the anisotropy field to
            match grid dimensions. Defaults to True.

        """
        # No Interpolation
        self._x = x_u[:,0].astype(dtype)
        self._y = x_u[:,1].astype(dtype)
        self._u = x_u[:,2].astype(dtype)
        self._v = x_u[:,3].astype(dtype)

        self.width = x_u[:, 0].max() - x_u[:, 0].min() if width is None \
                     else width

        self.height = x_u[:, 1].max() - x_u[:, 1].min() if height is None \
                     else height

        # Interpolated
        self.H = None
        self.V = None

        # Params
        self.beta = beta
        self.gamma = gamma
        self.dtype = dtype
        self.scale = scale
        self.k = k

    @property
    def x(self) -> np.ndarray:
        """Get the X-coordinates of vectors, considering interpolation.

        Returns
        -------
            np.ndarray: X-coordinates.

        """
        if self.V is None:
            return self._x
        else:
            j = np.indices(self.V.shape)[1,:,:,0]
            return j.reshape(-1)

    @property
    def y(self) -> np.ndarray:
        """Get the Y-coordinates of vectors, considering interpolation.

        Returns
        -------
            np.ndarray: Y-coordinates.

        """
        if self.V is None:
            return self._y
        else:
            i = np.indices(self.V.shape)[0,:,:,0]
            return i.reshape(-1)

    @property
    def u(self) -> np.ndarray:
        """Get the horizontal components of vectors, considering interpolation.

        Returns
        -------
            np.ndarray: Horizontal components.

        """
        if self.V is None:
            return self._u
        else:
            return self.V[:,:,1].reshape(-1)

    @property
    def v(self) -> np.ndarray:
        """Get the vertical components of vectors, considering interpolation.

        Returns
        -------
            np.ndarray: Vertical components.

        """
        if self.V is None:
            return self._v
        else:
            return self.V[:,:,0].reshape(-1)


    def plot(self, fig: Figure = None, ax: Axes = None, *, show: bool = True,
             ) -> Figure:
        """Plot the vectors.

        Args:
        ----
            fig (Figure, optional): Matplotlib Figure object to use for
            plotting. Defaults to None.
            ax (Axes, optional): Matplotlib Axes object to use for plotting.
            Defaults to None.
            show (bool, optional): Whether to display the plot. Defaults to
            True.

        Returns:
        -------
            Figure: Matplotlib Figure object.

        """
        if (fig is None) & (ax is None):
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
            ax.invert_yaxis()

        ax.quiver(self.x, self.y, self.u, self.v, angles='xy')

        if show:
            fig.show()

        return fig
    
    def check_positiveness(self):
        nx_ = self.H.shape[1]
        ny_ = self.H.shape[0]

        positiveness = np.zeros((ny_, nx_),dtype=bool)

        for j in range(nx_):
            for i in range(ny_):
                x = self.H[i, j, :, :]
                positiveness[i, j] = np.all(np.linalg.eigvals(x) > 0)

        return positiveness

    def check_differenciable(self, dx, dy):
        nx_ = self.H.shape[1]
        ny_ = self.H.shape[0]

        differenciable = np.zeros((ny_, nx_, 2, 2))

        for j in range(1,nx_-1):
            for i in range(1, ny_ - 1):

                dh_dy = (self.H[i + 1, j, :, :] - self.H[i - 1, j, :, :]) / (2 * dy)
                dh_dx = (self.H[i, j + 1, :, :] - self.H[i , j - 1, :, :]) / (2 * dx)

                dh00_dy = dh_dy[0, 0]
                dh01_dy = dh_dy[0, 1]
                dh11_dx = dh_dx[1, 1]
                dh10_dx = dh_dx[1, 0]


                differenciable[i, j, 0, 0] = dh00_dy
                differenciable[i, j, 0, 1] = dh01_dy
                differenciable[i, j, 1, 1] = dh11_dx
                differenciable[i, j, 1, 0] = dh10_dx

        return differenciable

def anisotropy_from_svg(path: str, beta: float, gamma: float,
                        norm_type: str = 'norm', k: int = 3) -> Anisotropy:
    """Create an Anisotropy object from an SVG file.

    Args:
    ----
        path (str): Path to the SVG file.
        beta (float): Beta parameter.
        gamma (float): Gamma parameter.
        norm_type (str, optional): Type of normalization. Defaults to 'norm'.

    Returns:
    -------
        Anisotropy: An Anisotropy object.

    """
    tree = Et.parse(path)
    root = tree.getroot()

    xmlns = get_inkscape_namespace(root)
    tfm = get_inkscape_transform(root, xmlns)


    width = float(root.attrib['width'][:-2])
    height = float(root.attrib['height'][:-2])

    P = []
    
    # Remove defitinions to skip style 'path' elements, like arrows heads.
    defs = next(root.iter(f'{xmlns}defs'))  
    root.remove(defs)

    for child in root.iter(f'{xmlns}path'):

        values = child.attrib['d'][2:]
        style = child.attrib['d'][0]

        p0, p1 = values.split(' ')

        p0 = p0.split(',')
        p1 = p1.split(',')

        p0 = [float(p0[0]), float(p0[1])]
        p1 = [float(p1[0]) , float(p1[1])]

        # m style is incremental / M style is absolute
        if style == 'M':
            p1 = [p1[0] - p0[0], p1[1] - p0[1]]

        # Flip anchor point (p0) and incremental point (p1)
        p0[1] =  p0[1]
        p1[1] = p1[1]

        ## When in quadrant III and IV, switch to quadrant I or II.
        #if p1[0] < 0:
        #    p1[0] *= -1
        #    p1[1] *= -1

        P.append([p0[0], p0[1], p1[0], p1[1]])

    P = np.asarray(P)

    x = P[:,:2] +  tfm
    u = P[:,2:]

    norm = (u**2).sum(axis=1)**0.5
    if norm_type == 'min':
        min_norm = (norm).min()
        u = u/(min_norm)
    elif norm_type == 'max':
        max_norm = (norm).max()
        u = u/(max_norm)
    elif norm_type == 'norm':
        u = u/norm[:, np.newaxis]
    else:
        msg = f'Unknown norm_type = "{norm_type}"'
        raise ValueError(msg)
    
    x_u = np.column_stack((x,u))

    return Anisotropy(x_u, beta, gamma, width,
                      height, k)


def anisotropy_stationary(theta, gamma, beta):

    theta = np.radians(theta)

    v =  np.asarray([- np.sin(theta),  np.cos(theta)])
    x_u  = np.asarray([[0,0,v[0],v[1]]])

    return Anisotropy(x_u, beta, gamma, scale=False)



from .utils import imdict
from dataclasses import dataclass
from typing import Callable

class Neighbours:
    """Neighbours Class.

    Class containing the flatten index of the 8 neighbouring elements in
    the 2d grid.

    Attributes
    ----------
    i : int
        The row index of the cell.
    j : int
        The column index of the cell.
    func : function
        A function used to calculate values for neighboring cells.
    allowable_neighs : list
        List of allowable neighboring cell indices.

    Methods
    -------
    __repr__()
        Returns a string representation of the neighboring cell indices and XY
        coordinates.

    """

    _neighs = imdict(id=imdict(i=0, j=0),
                     id_left=imdict(i=0, j=-1),
                     id_top_left=imdict(i=-1, j=-1),
                     id_top=imdict(i=-1, j=0),
                     id_top_right=imdict(i=-1, j=1),
                     id_right=imdict(i=0, j=1),
                     id_bot_right=imdict(i=1, j=1),
                     id_bot=imdict(i=1, j=0),
                     id_bot_left=imdict(i=1, j=-1))

    def __init__(self, i: int, j: int, func: Callable) -> None:
        """Initialize the Neighbours class.

        Parameters
        ----------
        i : int
            Row index.
        j : int
            Column index.
        func : Callable
            Index mapping function.

        """
        self.i = i
        self.j = j

        for pos, delta in self._neighs.items():
            setattr(self, pos, func(i + delta['i'], j + delta['j']))

        self.allowable_neighs = [k for k, v in self.__dict__.items()
                                 if k.startswith('id_') and v is not None]

    def has(self, pos: str) -> bool:
        """Check if neighbour exists.

        Parameters
        ----------
        pos : str
            Neighbour to check.

        Returns
        -------
        bool
            A boolean indicating whether the neighbour exists.

        """
        return getattr(self, pos, None) is not None



    def __repr__(self) -> str:
        """Return a schematic view of neighbours.

        For both 1D Flatten index and 2D XY Coordinates.
        """
        neighs = {}
        for neigh in self._neighs:
            value = getattr(self, neigh)
            if value is None:
                neighs[neigh] =' '*6
            else:
                neighs[neigh] = f'{value:06d}'

        ids = \
        "1D Flatten index:\n"+\
        "_"*27 + "\n"+\
        f"|{neighs['id_top_left']} | {neighs['id_top']} | {neighs['id_top_right']}|\n"+\
        f"|{neighs['id_left']} | {neighs['id']} | {neighs['id_right']}|\n"+\
        f"|{neighs['id_bot_left']} | {neighs['id_bot']} | {neighs['id_bot_right']}|\n"+\
        "‾"*27 + "\n\n"


        neighs = {}
        for neigh, delta in self._neighs.items():
            value = getattr(self, neigh)
            if value is None:
                neighs[neigh] =' '*11
            else:
                neighs[neigh] = f"({self.j + delta['j']:>4},"\
                                f"{self.i + delta['i']:>4})"

        xy = \
        "2D XY Coordinates:\n"+\
        "_"*41 + "\n"+\
        f"|{neighs['id_top_left']} | {neighs['id_top']} | {neighs['id_top_right']}|\n"+\
        f"|{neighs['id_left']} | {neighs['id']} | {neighs['id_right']}|\n"+\
        f"|{neighs['id_bot_left']} | {neighs['id_bot']} | {neighs['id_bot_right']}|\n"+\
        "‾"*41 + "\n"

        return ids + xy




from __future__ import annotations
from typing import TYPE_CHECKING
import warnings

import numpy as np
from  scipy.spatial import  KDTree

if TYPE_CHECKING:
    from .anisotropy import Anisotropy

__all__ = ['Grid']

class Grid:
    """Grid class representing a computational grid for anisotropic simulations.

    Attributes
    ----------
        nx (int): Number of grid points in the x-direction.
        ny (int): Number of grid points in the y-direction.
        dx (float): Spacing between grid points in the x-direction.
        dy (float): Spacing between grid points in the y-direction.
        anisotropy (Anisotropy): An Anisotropy object representing the
        anisotropy field.
        padx (int): Padding in the x-direction.
        pady (int): Padding in the y-direction.
        _nx (int): Total number of grid points including padding in the
        x-direction.
        _ny (int): Total number of grid points including padding in the
        y-direction.

    """

    _padx_pct = 15
    _pady_pct = 15

    def __init__(self, nx: int, ny: int, anisotropy: Anisotropy, dx: float=1,
                 dy: float=1, padx: None | int = None, pady: None | int = None,
                 ) -> None:
        """Initialize a Grid object.

        Args:
        ----
            nx (int): Number of grid points in the x-direction.
            ny (int): Number of grid points in the y-direction.
            dx (float): Spacing between grid points in the x-direction.
            dy (float): Spacing between grid points in the y-direction.
            anisotropy (Anisotropy): An Anisotropy object representing the
            anisotropy field.
            padx (None | int, optional): Padding in the x-direction. Defaults
            to None.
            pady (None | int, optional): Padding in the y-direction. Defaults
            to None.

        """
        self.nx = nx
        self.ny = ny
        self.dx = dx
        self.dy = dy
        self.anisotropy = anisotropy

        # Add Padding
        if padx is None:
            self.padx = int(np.ceil(self._padx_pct / 100 * nx))
        else:
            self.padx = padx

        if pady is None:
            self.pady = int(np.ceil(self._pady_pct / 100 * ny))
        else:
            self.pady = pady

        self._nx = nx + 2 * self.padx
        self._ny = ny + 2 * self.pady

        self.intrp_anisotropy()


    def intrp_anisotropy(self) ->  None:
        """Interpolate the anisotropy field onto the grid.

        Args:
        ----
            k (int, optional): Number of nearest neighbors to consider in
            interpolation. Defaults to 3.


        """
        anis = self.anisotropy
        k = min(anis.k,  len(anis.x))

        if anis.scale:
            scale_x = self.nx * self.dx / anis.width
            scale_y = self.ny * self.dy / anis.height

            # TODO(ejimenez): scale u,v
        else:
            scale_x = 1
            scale_y = 1

        x = (anis.x * scale_x) + self.padx
        y = (anis.y * scale_y) + self.pady

        X = np.column_stack((x, y))
        U = np.column_stack((anis.u, anis.v))

        # 1D
        grid = np.meshgrid(np.arange(self._nx) * self.dx,
                           np.arange(self._ny) * self.dy)

        # 2D
        grid = np.column_stack((grid[0].reshape(-1), grid[1].reshape(-1)))

        V = np.zeros(grid.shape, dtype=anis.dtype)
        tree = KDTree(X)

        dists, idxs = tree.query(grid, k=k)

        if k == 1:
            dists = dists[:, np.newaxis]
            idxs = idxs[:, np.newaxis]

        warnings.filterwarnings("error")

        for i,(dist, idx) in enumerate(zip(dists, idxs)):

            try:
                inv_dist = dist**(-0.25)
                inv_dist = inv_dist/inv_dist.sum()
            except RuntimeWarning:

                inv_dist = np.zeros(k, dtype=anis.dtype)
                inv_dist[np.argmin(dist)] = 1 


            V[i, :] = (U[idx, :] * inv_dist[:, np.newaxis]).sum(axis=0)

        warnings.resetwarnings()

        V = V.reshape(self._ny, self._nx , 2)[:,:,::-1]
        H = np.zeros((self._ny,self._nx, 2, 2), dtype=anis.dtype)
        I = np.eye(2, dtype=anis.dtype)

        for i in range(self._ny):
            for j in range(self._nx):
                H[i, j, :, :] = anis.gamma*I +\
                                anis.beta*(V[i, j][:,np.newaxis] @ \
                                           V[i, j][np.newaxis,:] )

        anis.H = H
        anis.V = V





from __future__ import annotations
import re

import numpy as np

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import xml.etree.ElementTree as Et

__all__ = ['imdict', 'map_2d_1d_nx_ny', 'map_1d_2d_nx', 'sample_grid',
           'get_inkscape_transform', 'get_inkscape_namespace']

class imdict(dict):
    def __hash__(self) -> None:
        """Return the unique id of the object as its hash value."""
        return id(self)

    def _immutable(self, *args, **kws) -> None:
        """Raise TypeError for methods that attempt mutability."""
        msg = 'Object is immutable'
        raise TypeError(msg)

    __setitem__ = _immutable
    __delitem__ = _immutable
    clear       = _immutable
    update      = _immutable
    setdefault  = _immutable
    pop         = _immutable
    popitem     = _immutable


def map_2d_1d_nx_ny(i: int, j: int, nx: int, ny: int) -> int | None:
    """Map a 2D index to a 1D index given the dimensions of the 2D array.

    Parameters
    ----------
    i : int
        The row index.
    j : int
        The column index.
    nx : int
        The total width of the 2D array.
    ny : int
        The total height of the 2D array.

    Returns
    -------
    int or None
        The mapped 1D index if the input indices are within bounds, otherwise
        None.

    """
    if (i < 0) or (j < 0) or (j >= nx) or (i >= ny):
        return None

    return int(nx*i + j)


def map_1d_2d_nx(idx: int, nx: int) -> tuple[int, int]:
    """Map a 1D index to a 2D index given the total width of the 2D array.

    Parameters
    ----------
    idx : int
        The 1D index to be mapped.
    nx : int
        The total width of the 2D array.

    Returns
    -------
    Tuple[int, int]
        A tuple containing the mapped 2D indices (i, j).

    """
    j = int(idx % nx)
    i = int((idx - j) / nx)

    return i, j


def get_inkscape_transform(root: Et.Element, xmlns: str,
                           ) -> tuple[float, float]:
    """Get the inkscape transform from the root element.

    Parameters
    ----------
    root : Et.Element
        The root element.
    xmlns : str
        The namespace string.

    Returns
    -------
    Tuple[float, float]
        A tuple containing the x and y shift values.

    """
    try:
        translate = next(root.iter(f'{xmlns}g')).attrib['transform']
        translate = re.search(r'translate\((\d+\.\d+),(\d+\.\d+)\)',translate)

        if translate is None:
            translate = (0, 0)
        else:
            shift_x = float(translate.group(1))
            shift_y = float(translate.group(2))

            translate = (shift_x, shift_y)

    except Exception as e:
        print('Unknown error while getting inkscape transform: ', e)
        translate = (0, 0)

    return translate


def get_inkscape_namespace(root: Et.Element) -> str:
    """Extract the XML namespace from the root element tag.

    Parameters
    ----------
    root : ElementTree.Element
        The root element of the XML document.

    Returns
    -------
    str
        The XML namespace extracted from the root element tag. If no namespace 
        is found, an empty string is returned.

    """
    xmlns = re.search(r'({.*})',root.tag)
    xmlns = xmlns.group(1) if xmlns is not None else ''

    return xmlns


def sample_grid(Z: np.ndarray, dx: float, dy: float,
             ) -> np.ndarray:
    """Regular sampling from a 2D grid.

    Parameters
    ----------
    Z : np.ndarray
        The input array from which to sample points.
    dx : float
        The spacing between sample points in the x-direction.
    dy : float
        The spacing between sample points in the y-direction.

    Returns
    -------
    np.ndarray
        An array of sample points with shape (n, 3), where n is the number of
        sample points.

    """
    ny, nx, * _ = Z.shape

    sx = np.arange(0, nx, dx, dtype=int)
    sy = np.arange(0, ny, dy, dtype=int)

    xv, yv = np.meshgrid(sx, sy, indexing='xy')

    xv = xv.reshape(-1)
    yv = yv.reshape(-1)
    samps = Z[yv, xv]

    X = np.column_stack((xv, yv, samps))

    return X


def append_element(vals: list, rows: list, cols: list, val: float, row: int,
                   col: int) -> None:
    """Append an element for cosntructiong the sparse matrix.

    Parameters
    ----------
    vals : list
        The list to append the value to.
    rows : list
        The list to append the row index to.
    cols : list
        The list to append the column index to.
    val : float
        The value to append to the `vals` list.
    row : int
        The row index to append to the `rows` list.
    col : int
        The column index to append to the `cols` list.

    Returns
    -------
    None

    """
    vals.append(val)
    rows.append(row)
    cols.append(col)
